/***************************************************************************
 *   Copyright (C) 2021 by Stefan Kebekus                                  *
 *   stefan.kebekus@gmail.com                                              *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 3 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#pragma once

#include <QGeoPositionInfo>

#include "traffic/TrafficDataSource_Abstract.h"


namespace Traffic {

/*! \brief Traffic receiver: Simulator that provides constant data
 *
 *  For testing purposes, this class provides constant traffic data.
 */
class TrafficDataSource_Simulate : public TrafficDataSource_Abstract {
    Q_OBJECT

public:
    /*! \brief Default constructor
     *
     *  @param parent The standard QObject parent pointer
     */
    explicit TrafficDataSource_Simulate(QObject *parent = nullptr);

    // Standard destructor
    ~TrafficDataSource_Simulate() override = default;

    /*! \brief Getter function for the property with the same name
     *
     *  This method implements the pure virtual method declared by its
     *  superclass.
     *
     *  @returns Property sourceName
     */
    QString sourceName() const override
    {
        return tr("Simulator data");
    }


public slots:
    /*! \brief Start attempt to connect to traffic receiver
     *
     *  This method implements the pure virtual method declared by its
     *  superclass.
     */
    void connectToTrafficReceiver() override;

    /*! \brief Disconnect from traffic receiver
     *
     *  This method implements the pure virtual method declared by its
     *  superclass.
     */
    void disconnectFromTrafficReceiver() override;

    /*! \brief Set coordinate that is to be reported by this class as the position of ownship
     *
     *  @param coordinate Coordinate of simulated ownship
     */
    void setCoordinate(const QGeoCoordinate& coordinate)
    {
        geoInfo.setCoordinate(coordinate);
    }

    /*! \brief Set angle that is to be reported by this class as the true track of ownship
     *
     *  @param TT True track of simulated ownship
     */
    void setTT(AviationUnits::Angle TT)
    {
        geoInfo.setAttribute(QGeoPositionInfo::Direction, TT.toDEG());
    }

    /*! \brief Set speed that is to be reported by this class as the ground speed of ownship
     *
     *  @param GS Ground speed of simulated ownship
     */
    void setGS(AviationUnits::Speed GS)
    {
        geoInfo.setAttribute(QGeoPositionInfo::GroundSpeed, GS.toMPS());
    }

    /*! \brief Set distance that is to be reported by this class as the barometric altitude of ownship
     *
     *  @param barAlt Barometric altitude of simulated ownship
     */
    void setBarometricHeight(AviationUnits::Distance barAlt)
    {
        barometricHeight = barAlt;
    }

private slots:
    // Send out simulated data. This slot will be called once per second once connectToTrafficReceiver() has been called
    void sendSimulatorData();

private:

    // Simulator related members
    QTimer simulatorTimer;
    QGeoPositionInfo geoInfo;
    AviationUnits::Distance barometricHeight;
};

}
